/*
    Copyright (c) 2002, 2003, bDistributed.com, Inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

    *   Redistributions of source code must retain the above
        copyright notice, this list of conditions and the following
        disclaimer.
    
    *   Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials
        provided with the distribution.
    
    *   Neither the name of bDistributed.com, Inc. nor the names of
        its contributors may be used to endorse or promote products
        derived from this software without specific prior written
        permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
    OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
    OF SUCH DAMAGE.
*/

#import "BDKeyComparisonQualifier.h"


@implementation BDKeyComparisonQualifier

- (id)initWithLeftKey:(NSString *)leftKey operatorSelector:(SEL)selector
             rightKey:(NSString *)rightKey
{
    if (self = [super init]) {
        _leftKey = [leftKey copy];
        _selector = selector;
        _rightKey = [rightKey copy];
    }

    return self;
}

- (void)dealloc
{
    [_leftKey release];
    [_rightKey release];
    [super dealloc];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"%@ %@ %@",
        _leftKey,
        [BDQualifier stringForOperatorSelector:_selector],
        _rightKey];
}

- (BOOL)evaluateWithObject:(id)object
{
    BOOL result = NO;

    if (object != nil) {
        id leftAttribute = nil;
        id rightAttribute = nil;
        NSMethodSignature *signature;
        NSInvocation *invocation;

        // First, get the attributes or relationships with the given left and right
        // keys from the object.  We call these the "left attribute" and "right
        // attribute" respectively.

        leftAttribute = [object valueForKey:_leftKey];
        rightAttribute = [object valueForKey:_rightKey];

        // Next, see if the left attribute responds to the comparison selector.

        if ([leftAttribute respondsToSelector:_selector] == NO) {
            return NO;
        }

        // Finally, send the left attribute the comparison with the right attribute
        // and use the result of that as our result.

        // Get the method signature for the comparison selector on the left attribute.

        signature = [leftAttribute methodSignatureForSelector:_selector];

        // Create an invocation using that method signature.

        invocation = [NSInvocation invocationWithMethodSignature:signature];

        // Set its first real argument to be the value to compare with.
        // (Arguments 0 and 1 are self and _cmd, respectively.)

        [invocation setArgument:&rightAttribute atIndex:2];

        // Invoke the invocation!

        [invocation invoke];

        // Get a return value out of the invocation.

        [invocation getReturnValue:&result];

        // And we're done!  We didn't even have to use a function pointer
        // (IMP).  Isn't introspection cool?
    }

    return result;
}

- (void)addQualifierKeysToSet:(NSMutableSet *)qualKeys
{
    [qualKeys addObject:_leftKey];
    [qualKeys addObject:_rightKey];
}

- (NSString *)leftKey
{
    return [[_leftKey retain] autorelease];
}

- (SEL)selector
{
    return _selector;
}

- (NSString *)rightKey
{
    return [[_rightKey retain] autorelease];
}

/* NSCoding Protocol */

- (id)initWithCoder:(NSCoder *)coder
{
    return [self initWithLeftKey:[coder decodeObject]
                operatorSelector:NSSelectorFromString([coder decodeObject])
                        rightKey:[coder decodeObject]];
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    NSString *selectorString = NSStringFromSelector(_selector);
    
    [coder encodeObject:_leftKey];
    [coder encodeObject:selectorString];
    [coder encodeObject:_rightKey];
}

@end
