/*
    Copyright (c) 2002, 2003, bDistributed.com, Inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

    *   Redistributions of source code must retain the above
        copyright notice, this list of conditions and the following
        disclaimer.
    
    *   Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials
        provided with the distribution.
    
    *   Neither the name of bDistributed.com, Inc. nor the names of
        its contributors may be used to endorse or promote products
        derived from this software without specific prior written
        permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
    OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
    OF SUCH DAMAGE.
*/

#import "BDKeyValueQualifier.h"
#import "BDQualifierVariable.h"


@implementation BDKeyValueQualifier

- (id)initWithKey:(NSString *)key operatorSelector:(SEL)selector value:(id)value
{
    if (self = [super init]) {
        _key = [key copy];
        _selector = selector;
        _value = [value retain];
    }

    return self;
}

- (void)dealloc
{
    [_key release];
    [_value release];
    [super dealloc];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"%@ %@ %@",
        _key,
        [BDQualifier stringForOperatorSelector:_selector],
        [self formatValueBasedOnType:_value]];
}

- (BOOL)evaluateWithObject:(id)object
{
    BOOL result = NO;

    if (object != nil) {
        id target = nil;
        NSMethodSignature *signature;
        NSInvocation *invocation;

        // First, get the attribute or relationship with the given key from the
        // object.  We call this the target.

        target = [object valueForKeyPath:_key];

        if (target != nil) {
            // Next, see if the target responds to the comparison selector.

            if ([target respondsToSelector:_selector] == NO) {
                return NO;
            }

            // Finally, send the target the comparison with the value and use the result
            // of that as our result.

            // Get the method signature for the comparison selector on the target.

            signature = [target methodSignatureForSelector:_selector];

            // Create an invocation using that method signature and selector.

            invocation = [NSInvocation invocationWithMethodSignature:signature];

            [invocation setSelector:_selector];

            // Set its first real argument to be the value to compare with.
            // (Arguments 0 and 1 are self and _cmd, respectively.)

            [invocation setArgument:&_value atIndex:2];

            // Invoke the invocation!

            [invocation invokeWithTarget:target];

            // Get a return value out of the invocation.

            [invocation getReturnValue:&result];

            // And we're done!  We didn't even have to use a function pointer
            // (IMP).  Isn't introspection cool?
        } else {
            // The target is nil.  This is valid only if the comparison is
            // isEqualTo: or isNotEqualTo:.  If it's not one of those, the
            // result of the comparison will be NO.

            if (_selector == BDQualifierOperatorEqual) {
                // We're checking for equality.

                if ((_value == nil) || (_value == [NSNull null])) {
                    result = YES;
                } else {
                    result = NO;
                }
            } else if (_selector == BDQualifierOperatorNotEqual) {
                // We're checking for inequality.

                if ((_value == nil) || (_value == [NSNull null])) {
                    result = NO;
                } else {
                    result = YES;
                }
            } else {
                result = NO;
            }
        }
    }

    return result;
}

- (void)addQualifierKeysToSet:(NSMutableSet *)qualKeys
{
    [qualKeys addObject:_key];
}

- (int)minimumQualifiers
{
    return 0;
}

- (int)maximumQualifiers
{
    return 0;
}

- (NSString *)key
{
    return [[_key retain] autorelease];
}

- (SEL)selector
{
    return _selector;
}

- (id)value
{
    return [[_value retain] autorelease];
}

/* Accessing a qualifier's binding keys */

- (NSArray *)bindingKeys
{
    NSArray *bindingKeys;
    
    // If our value is an instance of BDQualifierVariable, it
    // is actually a binding key.  Return it if it is; otherwise,
    // return an empty array.

    if ([_value isKindOfClass:[BDQualifierVariable class]]) {
        bindingKeys = [NSArray arrayWithObject:[_value key]];
    } else {
        bindingKeys = [NSArray array];
    }

    return bindingKeys;
}

- (NSString *)keyPathForBindingKey:(NSString *)key
{
    NSString *keyPath = nil;

    if ([_value isKindOfClass:[BDQualifierVariable class]]) {
        if ([[_value key] isEqualToString:key]) {
            keyPath = [self key];
        }
    }

    return keyPath;
}

- (BDQualifier *)qualifierWithBindings:(NSDictionary *)bindings
                  requiresAllVariables:(BOOL)requiresAll
{
    BDQualifier *newQualifier = nil;
    NSString *bindingKey;
    id newValue;

    if ([_value isKindOfClass:[BDQualifierVariable class]]) {
        bindingKey = [_value key];

        newValue = [bindings objectForKey:bindingKey];

        if (newValue == nil) {
            if (requiresAll) {
                [NSException raise:BDQualifierVariableSubstitutionException
                            format:@"BDQualifierVariableSubstitutionException: Missing required binding '%@'",
                    bindingKey];
            } else {
                // Prune this qualifier
                
                newQualifier = nil;
            }
        } else {
            newQualifier = [[[BDKeyValueQualifier alloc] initWithKey:_key
                                                    operatorSelector:_selector
                                                               value:newValue] autorelease];
        }
    } else {
        newQualifier = self;
    }
    
    return newQualifier;
}

/* NSCoding Protocol */

- (id)initWithCoder:(NSCoder *)coder
{
    return [self initWithKey:[coder decodeObject]
            operatorSelector:NSSelectorFromString([coder decodeObject])
                       value:[coder decodeObject]];
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [coder encodeObject:_key];
    [coder encodeObject:NSStringFromSelector(_selector)];
    [coder encodeObject:_value];
}

@end
