/*
    Copyright (c) 2002, 2003, bDistributed.com, Inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

    *   Redistributions of source code must retain the above
        copyright notice, this list of conditions and the following
        disclaimer.
    
    *   Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials
        provided with the distribution.
    
    *   Neither the name of bDistributed.com, Inc. nor the names of
        its contributors may be used to endorse or promote products
        derived from this software without specific prior written
        permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
    OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
    OF SUCH DAMAGE.
*/

#import "BDOrQualifier.h"


@implementation BDOrQualifier

- (id)initWithQualifierArray:(NSArray *)qualifiers
{
    if (self = [super init]) {
        _qualifiers = [qualifiers copy];
    }

    return self;
}

- (id)initWithQualifiers:(BDQualifier *)qualifier, ...
{
    NSMutableArray *qualifiers;
    BDQualifier *arg;
    va_list ap;

    va_start(ap, qualifier);

    qualifiers = [[NSMutableArray alloc] initWithCapacity:2];
    [qualifiers addObject:qualifier];

    while (arg = va_arg(ap, BDQualifier *)) {
        [qualifiers addObject:arg];
    }

    va_end(ap);

    return [self initWithQualifierArray:qualifiers];
}

- (NSString *)description
{
    NSMutableString *d = [[NSMutableString alloc] init];
    BOOL firstTime = YES;
    NSEnumerator *enumerator = [_qualifiers objectEnumerator];
    BDQualifier *oneQualifier;

    while (oneQualifier = [enumerator nextObject]) {
        if (firstTime == YES) {
            firstTime = NO;
        } else {
            [d appendString:@" or "];
        }
        [d appendString:@"("];
        [d appendString:[oneQualifier description]];
        [d appendString:@")"];
    }

    return [d autorelease];
}

- (BOOL)evaluateWithObject:(id)object
{
    BOOL result = NO;

    if (object != nil) {
        NSEnumerator *enumerator = [_qualifiers objectEnumerator];
        BDQualifier *oneQualifier;

        // Go through the sub-qualifiers.  As long as one returns YES,
        // the entire qualifier should return YES.  Otherwise, return
        // NO.

        while (oneQualifier = [enumerator nextObject]) {
            if ([oneQualifier evaluateWithObject:object] == YES) {
                return YES;
            }
        }
        result = NO;
    }

    return result;
}

- (unsigned)count
{
    int count = 1;
    NSEnumerator *enumerator = [_qualifiers objectEnumerator];
    BDQualifier *oneQualifier;

    // Go through the sub-qualifiers, and ask each one how many
    // qualifiers it has.  Accumulate this quantity and return it.

    while (oneQualifier = [enumerator nextObject]) {
        count += [oneQualifier count];
    }

    return count;
}

- (NSArray *)subqualifiers
{
    return [[_qualifiers copy] autorelease];
}

- (void)addQualifierKeysToSet:(NSMutableSet *)qualKeys
{
    // Iterate over any subqualifiers and add their keys to the set.

    NSEnumerator *enumerator = [_qualifiers objectEnumerator];
    BDQualifier *oneQualifier;

    while (oneQualifier = [enumerator nextObject]) {
        [oneQualifier addQualifierKeysToSet:qualKeys];
    }
}

/* Accessing a qualifier's binding keys */

- (NSArray *)bindingKeys
{
    NSMutableSet *bindingKeys = nil;
    NSEnumerator *qualifiersEnumerator;
    BDQualifier *oneQualifier;

    // Create a set (since each binding key only appears once)

    bindingKeys = [[[NSMutableSet alloc] init] autorelease];

    // Loop over all the sub-qaulifiers

    qualifiersEnumerator = [_qualifiers objectEnumerator];

    while (oneQualifier = [qualifiersEnumerator nextObject]) {

        // Add the binding keys for each sub-qualifier to the set

        [bindingKeys addObjectsFromArray:[oneQualifier bindingKeys]];
    }

    // Return all of the objects in the set as an array

    return [bindingKeys allObjects];
}

- (NSString *)keyPathForBindingKey:(NSString *)key
{
    NSString *keyPath = nil;
    NSEnumerator *qualifiersEnumerator;
    BDQualifier *oneQualifier;

    // Loop over all the sub-qualifiers

    qualifiersEnumerator = [_qualifiers objectEnumerator];

    while (oneQualifier = [qualifiersEnumerator nextObject]) {

        // Ask each sub-qualifier in turn for the key path for the
        // given binding key

        keyPath = [oneQualifier keyPathForBindingKey:key];

        // If it actually returned one, break out and return that.
        // Otherwise, continue with the next sub-qualifier.

        if (keyPath != nil) {
            break;
        }
    }

    return keyPath;
}

- (BDQualifier *)qualifierWithBindings:(NSDictionary *)bindings
                  requiresAllVariables:(BOOL)requiresAll
{
    NSEnumerator *enumerator;
    BDQualifier *oneQualifier;
    BDQualifier *newQualifier;
    BDQualifier *result;
    NSMutableArray *newQualifiers;

    newQualifiers = [NSMutableArray arrayWithCapacity:[_qualifiers count]];

    enumerator = [_qualifiers objectEnumerator];

    while (oneQualifier = [enumerator nextObject]) {
        newQualifier = [oneQualifier qualifierWithBindings:bindings
                                      requiresAllVariables:requiresAll];

        if (newQualifier != nil) {
            [newQualifiers addObject:newQualifier];
        }
    }

    if ([newQualifiers count] > 1) {
        result = [[[BDOrQualifier alloc] initWithQualifierArray:newQualifiers] autorelease];
    } else if ([newQualifiers count] == 1) {
        result = [[[newQualifiers objectAtIndex:0] retain] autorelease];
    } else {
        result = nil;
    }

    return result;
}

/* NSCoding Protocol */

- (id)initWithCoder:(NSCoder *)coder
{
    NSArray *qualifiers;
    
    qualifiers = [coder decodeObject];
    
    return [self initWithQualifierArray:qualifiers];
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [coder encodeObject:_qualifiers];
}

@end
