/*
    Copyright (c) 2002, 2003, bDistributed.com, Inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

    *   Redistributions of source code must retain the above
        copyright notice, this list of conditions and the following
        disclaimer.
    
    *   Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials
        provided with the distribution.
    
    *   Neither the name of bDistributed.com, Inc. nor the names of
        its contributors may be used to endorse or promote products
        derived from this software without specific prior written
        permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
    OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
    OF SUCH DAMAGE.
*/

#import "BDQualifier.h"
#import "BDKeyValueQualifier.h"
#import "BDAndQualifier.h"
#import "BDOrQualifier.h"
#import "BDQualifierVariable.h"

#import "BDQualifierParserExternal.h"


NSString *BDQualifierVariableSubstitutionException = @"BDQualifierVariableSubstitutionException";


@implementation BDQualifier

+ (void)initialize
{
    BDQualifierParserInitialize();
}

- (id)init
{
    if (self = [super init]) {
        // Do nothing
    }

    return self;
}

- (void)dealloc
{
    [super dealloc];
}

+ (BDQualifier *)qualifierWithQualifierFormat:(NSString *)qualifierFormat, ...
{
    BDQualifier *qualifier = nil;
    va_list ap;

    va_start(ap, qualifierFormat);

    return [self qualifierWithQualifierFormat:qualifierFormat
                                   varargList:ap];

    va_end(ap);
    
    return qualifier;
}

+ (BDQualifier *)qualifierWithQualifierFormat:(NSString *)qualifierFormat
                                   varargList:(va_list)args
{
    BDQualifier *qualifier = nil;
    NSString *formattedString;

    // Fill out the qualifier format into a proper string

    formattedString = [[[NSString alloc] initWithFormat:qualifierFormat
                                              arguments:args] autorelease];
    
    // Parse it into a qualifier

    qualifier = BDQualifierParseCString([formattedString UTF8String]);

    return qualifier;
}

+ (BDQualifier *)qualifierToMatchAllValues:(NSDictionary *)values
{
    BDQualifier *andQualifier = nil;
    NSMutableArray *keyValueQualifiers;
    NSEnumerator *keyEnumerator;
    NSString *key;

    keyValueQualifiers = [NSMutableArray arrayWithCapacity:[values count]];
    
    keyEnumerator = [values keyEnumerator];
    
    while (key = [keyEnumerator nextObject]) {
        BDKeyValueQualifier *q;
        
        q = [[BDKeyValueQualifier alloc] initWithKey:key
                                    operatorSelector:BDQualifierOperatorEqual
                                               value:[values valueForKey:key]];
        
        [keyValueQualifiers addObject:[q autorelease]];
    }

    andQualifier = [[BDAndQualifier alloc] initWithQualifierArray:keyValueQualifiers];

    return [andQualifier autorelease];
}

+ (BDQualifier *)qualifierToMatchAnyValue:(NSDictionary *)values
{
    BDQualifier *orQualifier = nil;
    NSMutableArray *keyValueQualifiers;
    NSEnumerator *keyEnumerator;
    NSString *key;

    keyValueQualifiers = [NSMutableArray arrayWithCapacity:[values count]];
    keyEnumerator = [values keyEnumerator];

    while (key = [keyEnumerator nextObject]) {
        BDKeyValueQualifier *q;

        q = [[BDKeyValueQualifier alloc] initWithKey:key
                                    operatorSelector:BDQualifierOperatorEqual
                                               value:[values valueForKey:key]];
        
        [keyValueQualifiers addObject:[q autorelease]];
    }

    orQualifier = [[BDOrQualifier alloc] initWithQualifierArray:keyValueQualifiers];

    return [orQualifier autorelease];
}

- (BDQualifier *)qualifierWithBindings:(NSDictionary *)bindings
                  requiresAllVariables:(BOOL)requiresAll
{
    return self;
}

- (BOOL)evaluateWithObject:(id)object
{
    return NO;
}

- (unsigned)count
{
    // Return the number of sub-qualifiers in this qualifier plus one.
    // Subclasses that can have sub-qaulfiers should override this method.

    return 1;
}

- (NSArray *)subqualifiers
{
    // Return an array of the sub-qualifiers in this qualifier.  Return nil
    // if this qualifier can't have sub-qualifiers, and an empty array if it
    // can but doesn't.  Subclasses that can have sub-qaulfiers should override
    // this method.
    
    return nil;
}

+ (SEL)operatorSelectorForString:(NSString *)aString
{
    SEL sel;

    if ([aString isEqualToString:@"="] || [aString isEqualToString:@"=="]) {
        sel = BDQualifierOperatorEqual;
    } else if ([aString isEqualToString:@"!="]) {
        sel = BDQualifierOperatorNotEqual;
    } else if ([aString isEqualToString:@"<"]) {
        sel = BDQualifierOperatorLessThan;
    } else if ([aString isEqualToString:@"<"]) {
        sel = BDQualifierOperatorGreaterThan;
    } else if ([aString isEqualToString:@"<="]) {
        sel = BDQualifierOperatorLessThanOrEqualTo;
    } else if ([aString isEqualToString:@">="]) {
        sel = BDQualifierOperatorGreaterThanOrEqualTo;
    } else if ([aString isEqualToString:@"contains"]) {
        sel = BDQualifierOperatorContains;
    } else if ([aString isEqualToString:@"like"]) {
        sel = BDQualifierOperatorLike;
    } else if ([aString isEqualToString:@"caseInsensitiveLike"]) {
        sel = BDQualifierOperatorCaseInsensitiveLike;
    } else {
        sel = NULL;
    }

    return sel;
}

+ (NSString *)stringForOperatorSelector:(SEL)aSelector
{
    NSString *str = nil;

    if (aSelector == BDQualifierOperatorEqual) {
        str = @"=";
    } else if (aSelector == BDQualifierOperatorNotEqual) {
        str = @"!=";
    } else if (aSelector == BDQualifierOperatorLessThan) {
        str = @"<";
    } else if (aSelector == BDQualifierOperatorGreaterThan) {
        str = @">";
    } else if (aSelector == BDQualifierOperatorLessThanOrEqualTo) {
        str = @"<=";
    } else if (aSelector == BDQualifierOperatorGreaterThanOrEqualTo) {
        str = @">=";
    } else if (aSelector == BDQualifierOperatorContains) {
        str = @"contains";
    } else if (aSelector == BDQualifierOperatorLike) {
        str = @"like";
    } else if (aSelector == BDQualifierOperatorCaseInsensitiveLike) {
        str = @"caseInsensitiveLike";
    } else {
        str = nil;
    }

    return str;
}

+ (NSArray *)allQualifierOperators
{
    return [NSArray arrayWithObjects:@"=",
        @"!=",
        @"<",
        @">",
        @"<=",
        @">=",
        @"contains",
        @"like",
        @"caseInsensitiveLike",
        nil];
}

+ (NSArray *)relationalQualifierOperators
{
    return [NSArray arrayWithObjects:@"=",
        @"!=",
        @"<",
        @">",
        @"<=",
        @">=",
        nil];
}

- (NSSet *)allQualifierKeys
{
    NSMutableSet *set = [NSMutableSet setWithCapacity:2];

    [self addQualifierKeysToSet:set];

    return set;
}

- (void)addQualifierKeysToSet:(NSMutableSet *)qualKeys
{
    // Do nothing
}

- (NSArray *)bindingKeys
{
    return [NSArray array];
}

- (NSString *)keyPathForBindingKey:(NSString *)key
{
    return nil;
}

- (NSException *)validateKeysWithRootClassDescription:(NSClassDescription *)classDesc
{
    //xxx
    return nil;
}

/*
	Formats a given object based on its type.  Strings are bracketed with single
	quotes, NSNulls are represented by nil, BDQualifierVariables are a dollar sign
	prepended to their key, and everything else is just asked for its description.
	There's probably a better way to do this, but this'll do for now.
*/
- (NSString *)formatValueBasedOnType:(id)value
{
    if (value == nil) {
        return @"nil";
    } else if ([value isKindOfClass:[NSString class]]) {
        return [NSString stringWithFormat:@"'%@'", value];
    } else if ([value isKindOfClass:[NSNull class]]) {
        return @"nil";
    } else if ([value isKindOfClass:[BDQualifierVariable class]]) {
        return [NSString stringWithFormat:@"$%@", [value key]];
    }

    return [value description];
}

/* NSCopying */

- (id)copyWithZone:(NSZone *)zone
{
    // Contents are immutable, so just retain
    
    return [self retain];
}

@end


@implementation NSArray (BDQualifierAdditions)

- (NSArray *)filteredArrayUsingQualifier:(BDQualifier *)qualifier
{
    NSMutableArray *result = nil;

    if (qualifier != nil) {
        NSEnumerator *enumerator = [self objectEnumerator];
        id oneObject = nil;

        result = [NSMutableArray arrayWithCapacity:0];

        while (oneObject = [enumerator nextObject]) {
            if ([qualifier evaluateWithObject:oneObject]) {
                [result addObject:oneObject];
            }
        }
    }

    return result;
}

@end
