/*
    Copyright (c) 2002, 2003, bDistributed.com, Inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

    *   Redistributions of source code must retain the above
        copyright notice, this list of conditions and the following
        disclaimer.
    
    *   Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials
        provided with the distribution.
    
    *   Neither the name of bDistributed.com, Inc. nor the names of
        its contributors may be used to endorse or promote products
        derived from this software without specific prior written
        permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
    OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
    OF SUCH DAMAGE.
*/

#import <Foundation/Foundation.h>

#import "BDQualifierParserExternal.h"

#import "BDQualifier.h"
#import "BDAndQualifier.h"
#import "BDOrQualifier.h"
#import "BDNotQualifier.h"
#import "BDBooleanQualifier.h"
#import "BDKeyValueQualifier.h"
#import "BDKeyComparisonQualifier.h"
#import "BDQualifierVariable.h"

#include "BDQualifierNode.h"


extern int BDQualifierParser_parse(void);
extern BDQualifierNode *BDQualifierParser_lastParsedNode;
extern void BDQualifierParser_restart(FILE *newFile);
extern void BDQualifierParser_reset_state();

static NSLock *BDQualifierParserLock = nil;
static char *BDQualifierParserBuffer = NULL;
static char *BDQualifierParserBufferCursor = NULL;
static int BDQualifierParserBufferLength = NULL;


static BDQualifier *BDQualifierNodeToQualifier(BDQualifierNode *node);


inline int min(int a, int b)
{
    if (a < b) {
        return a;
    }
    return b;
}


int BDQualifierParser_YYINPUT(char *buf, int max_size)
{
    int n = min(max_size, strlen(BDQualifierParserBufferCursor));

    if (n > 0) {
        memcpy(buf, BDQualifierParserBufferCursor, n);
        BDQualifierParserBufferCursor += n;
    }

    return n;
}


void BDQualifierParserInitialize(void)
{
    if (BDQualifierParserLock == nil) {
        BDQualifierParserLock = [[NSLock alloc] init];
    }
}


void BDQualifierParserTerminate(void)
{
    if (BDQualifierParserLock != nil) {
        [BDQualifierParserLock release];
        BDQualifierParserLock = nil;
    }
}


BDQualifier *BDQualifierParseCString(const char *qualifierString)
{
    BDQualifier *qualifier = nil;

    if (qualifierString == NULL) {
        return nil;
    }

    // Acquire the lock

    [BDQualifierParserLock lock];

    NS_DURING
        // Reset the parser

        BDQualifierParser_reset_state();
        BDQualifierParser_restart(NULL);
        
        // Set up parsing the given C string

        BDQualifierParserBuffer = (char *)qualifierString;
        BDQualifierParserBufferCursor = BDQualifierParserBuffer;
        BDQualifierParserBufferLength = strlen(BDQualifierParserBuffer);

        // Call the parser

        BDQualifierParser_parse();

        // Convert the last parsed node to a qualifier object

        qualifier = BDQualifierNodeToQualifier(BDQualifierParser_lastParsedNode);
    NS_HANDLER
        // Release the lock

        [BDQualifierParserLock unlock];

        // Re-raise the exception

        [localException raise];
    NS_ENDHANDLER

    // Release the lock

    [BDQualifierParserLock unlock];

    // Return the qualifier

    return qualifier;
}


BDQualifier *BDQualifierNodeToQualifier(BDQualifierNode *node)
{
    BDQualifier *qualifier = nil;

    if (node != NULL) {
        switch (node->type) {
            case kBDAndQualifierNodeType:
            {
                BDQualifier *leftQual = nil;
                BDQualifier *rightQual = nil;

                leftQual = BDQualifierNodeToQualifier(node->data.andNode.leftQual);
                rightQual = BDQualifierNodeToQualifier(node->data.andNode.rightQual);

                qualifier = [[[BDAndQualifier alloc] initWithQualifiers:leftQual, rightQual, nil] autorelease];
            }
                break;

            case kBDOrQualifierNodeType:
            {
                BDQualifier *leftQual = nil;
                BDQualifier *rightQual = nil;

                leftQual = BDQualifierNodeToQualifier(node->data.orNode.leftQual);
                rightQual = BDQualifierNodeToQualifier(node->data.orNode.rightQual);

                qualifier = [[[BDOrQualifier alloc] initWithQualifiers:leftQual, rightQual, nil] autorelease];
            }
                break;

            case kBDNotQualifierNodeType:
            {
                BDQualifier *internalQual = nil;

                internalQual = BDQualifierNodeToQualifier(node->data.notNode.qual);

                qualifier = [[[BDNotQualifier alloc] initWithQualifier:internalQual] autorelease];
            }
                break;

            case kBDKeyValueQualifierNodeType:
            {
                NSString *key;
                NSString *operatorSelector;
                SEL sel;
                id value = nil;

                key = [NSString stringWithCString:node->data.keyValueNode.key];
                operatorSelector = [NSString stringWithCString:node->data.keyValueNode.operatorSelector];
                sel = NSSelectorFromString(operatorSelector);

                switch (node->data.keyValueNode.valueType) {
                    case kBDQualifierNodeValueNumber:
                        value = [NSNumber numberWithDouble:node->data.keyValueNode.valueData.doubleValue];
                        break;

                    case kBDQualifierNodeValueString:
                        value = [NSString stringWithCString:node->data.keyValueNode.valueData.stringValue];
                        break;

                    case kBDQualifierNodeValueNil:
                        value = [NSNull null];
                        break;

                    case kBDQualifierNodeValueTypedString:
                    {
                        NSString *valueClassName;
                        NSString *valueStr;
                        Class valueClass;

                        valueClassName = [NSString stringWithCString:node->data.keyValueNode.valueData.typedStringType];
                        valueStr = [NSString stringWithCString:node->data.keyValueNode.valueData.typedStringValue];

                        valueClass = NSClassFromString(valueClassName);

                        value = [[[valueClass alloc] initWithString:valueStr] autorelease];
                    }
                        break;

                    case kBDQualifierNodeValueVariable:
                    {
                        NSString *variableKey;

                        variableKey = [NSString stringWithCString:node->data.keyValueNode.valueData.variableValue];

                        value = [BDQualifierVariable variableWithKey:variableKey];
                    }
                        break;

                    case kBDQualifierNodeValueBoolean:
                    {
                        if (node->data.keyValueNode.valueData.booleanValue) {
                            value = [NSNumber numberWithBool:YES];
                        } else {
                            value = [NSNumber numberWithBool:NO];
                        }
                    }
                        break;

                    default:
                        // Do nothing, let it be nil
                        break;

                }

                qualifier = [[[BDKeyValueQualifier alloc] initWithKey:key
                                                     operatorSelector:sel
                                                                value:value] autorelease];
            }
                break;

            case kBDKeyComparisonQualifierNodeType:
            {
                NSString *leftKey;
                NSString *operatorSelector;
                NSString *rightKey;
                SEL sel;

                leftKey = [NSString stringWithCString:node->data.keyComparisonNode.leftKey];
                operatorSelector = [NSString stringWithCString:node->data.keyComparisonNode.operatorSelector];
                sel = NSSelectorFromString(operatorSelector);
                rightKey = [NSString stringWithCString:node->data.keyComparisonNode.rightKey];

                qualifier = [[[BDKeyComparisonQualifier alloc] initWithLeftKey:leftKey
                                                              operatorSelector:sel
                                                                      rightKey:rightKey] autorelease];
            }
                break;

            case kBDBooleanQualifierNodeType:
                if (node->data.booleanNode == 0) {
                    qualifier = [[[BDBooleanQualifier alloc] initWithValue:NO] autorelease];
                } else {
                    qualifier = [[[BDBooleanQualifier alloc] initWithValue:YES] autorelease];
                }
                break;
                
            default:
                // Do nothing
                break;
        }
    }

    return qualifier;
}
