/*
    Copyright (c) 2002, 2003, bDistributed.com, Inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

    *   Redistributions of source code must retain the above
        copyright notice, this list of conditions and the following
        disclaimer.
    
    *   Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials
        provided with the distribution.
    
    *   Neither the name of bDistributed.com, Inc. nor the names of
        its contributors may be used to endorse or promote products
        derived from this software without specific prior written
        permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
    OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
    OF SUCH DAMAGE.
*/

#import "BDSortOrdering.h"


static int mergesortComparison(const void *a, const void *b);


@implementation BDSortOrdering

- (id)initWithKey:(NSString *)key
         selector:(SEL)selector
{
    if (self = [super init]) {
        _key = [key copy];
        _selector = selector;
    }
    return self;
}

- (void)dealloc
{
    [_key release];
    [super dealloc];
}

- (NSString *)key
{
    return [[_key retain] autorelease];
}

- (SEL)selector
{
    return _selector;
}

+ (BDSortOrdering *)sortOrderingWithKey:(NSString *)key
                               selector:(SEL)selector
{
    return [[[BDSortOrdering alloc] initWithKey:key
                                       selector:selector] autorelease];
}

// NSCoding Protocol

- (id)initWithCoder:(NSCoder *)coder
{
    return [self initWithKey:[coder decodeObject]
                    selector:NSSelectorFromString([coder decodeObject])];
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [coder encodeObject:_key];
    [coder encodeObject:NSStringFromSelector(_selector)];
}

@end

@interface BDSortOrdering (MergeSortImplementation)

+ (BOOL)mergeSortObjects:(id *)objects
                   count:(unsigned)count
                ordering:(BDSortOrdering *)ordering;

+ (void)mergesortArrayInPlace:(NSMutableArray *)array
           usingKeyOrderArray:(NSArray *)keyOrderArray;

+ (NSArray *)mergesortedArrayFromArray:(NSArray *)array
                    usingKeyOrderArray:(NSArray *)keyOrderArray;

@end

@implementation BDSortOrdering (MergeSortImplementation)

+ (BOOL)mergeSortObjects:(id *)objects
                   count:(unsigned)count
                ordering:(BDSortOrdering *)ordering
{
    BOOL success;
    int i;

    // Set the ordering for each object to the ordering passed in.
    // This hack is necessary because mergesort won't let us pass
    // a refcon.  Objects are actually (object, ordering) pairs.

    for (i = 0; i < count; i++) {
        objects[(i * 2) + 1] = ordering;
    }

    // Call mergesort to sort objects in-place using our generic
    // comparison function.

    if (mergesort(objects, count, (sizeof(id) * 2), mergesortComparison) == 0) {
        success = YES;
    } else {
        success = NO;
    }

    return success;
}

+ (void)mergesortArrayInPlace:(NSMutableArray *)array
           usingKeyOrderArray:(NSArray *)keyOrderArray
{
    int count;
    NSMutableData *objectsStorage;
    id *objects;

    // Construct a C array of (object, ordering) pairs for the input array.

    count = [array count];
    
    objectsStorage = [NSMutableData dataWithCapacity:([array count] * (sizeof(id) * 2))];
    objects = [objectsStorage mutableBytes];

    if (objects != NULL) {
        NSEnumerator *arrayEnumerator;
        id object;
        int i;

        arrayEnumerator = [array objectEnumerator];
        i = 0;

        while (object = [arrayEnumerator nextObject]) {
            objects[(i * 2) + 0] = object;

            i++;
        }
    }

    // Now go through the key-order array in reverse order and do a merge sort for each one.

    {
        BOOL success = YES;
        NSEnumerator *orderingEnumerator;
        BDSortOrdering *ordering;
        int j;

        orderingEnumerator = [keyOrderArray reverseObjectEnumerator];

        while (ordering = [orderingEnumerator nextObject]) {
            success = [self mergeSortObjects:objects
                                       count:count
                                    ordering:ordering];

            if (success == NO) {
                [NSException raise:NSMallocException
                            format:@"mergesort(3) failed"];
            }
        }

        // Copy back to the result array.
        
        for (j = 0; j < count; j++) {
            id sortedObject = objects[(j * 2) + 0];
            
            [array replaceObjectAtIndex:j
                             withObject:sortedObject];
        }
    }
}

+ (NSArray *)mergesortedArrayFromArray:(NSArray *)array
                    usingKeyOrderArray:(NSArray *)keyOrderArray
{
    NSMutableArray *resultArray = nil;
    int count;
    NSMutableData *objectsStorage;
    id *objects;
    
    // Construct a C array of (object, ordering) pairs for the input array.

    count = [array count];
    
    objectsStorage = [NSMutableData dataWithCapacity:([array count] * (sizeof(id) * 2))];
    objects = [objectsStorage mutableBytes];

    if (objects != NULL) {
        NSEnumerator *arrayEnumerator;
        id object;
        int i;

        arrayEnumerator = [array objectEnumerator];
        i = 0;

        while (object = [arrayEnumerator nextObject]) {
            objects[(i * 2) + 0] = object;

            i++;
        }
    }

    // Now go through the key-order array in reverse order and do a merge sort for each one.

    {
        BOOL success = YES;
        NSEnumerator *orderingEnumerator;
        BDSortOrdering *ordering;
        int j;

        orderingEnumerator = [keyOrderArray reverseObjectEnumerator];
    
        while (ordering = [orderingEnumerator nextObject]) {
            success = [self mergeSortObjects:objects
                                       count:count
                                    ordering:ordering];

            if (success == NO) {
                [NSException raise:NSMallocException
                            format:@"mergesort(3) failed"];
            }
        }
        
        // Construct the result array.

        resultArray = [NSMutableArray arrayWithCapacity:count];

        for (j = 0; j < count; j++) {
            id sortedObject = objects[(j * 2) + 0];

            [resultArray addObject:sortedObject];
        }
    }
    
    return resultArray;
}

@end


static int mergesortComparison(const void *inA, const void *inB)
{
    id *aBase = (id *)inA;
    id *bBase = (id *)inB;
    id a = aBase[0];
    id b = bBase[0];
    BDSortOrdering *ordering = aBase[1]; // b[1] is the same thing
    NSString *key = [ordering key];
    SEL selector = [ordering selector];
    int result;

    result = (int) [[a valueForKeyPath:key] performSelector:selector
                                                 withObject:[b valueForKeyPath:key]];

    return result;
}


@implementation NSObject (BDSortOrderingAdditions)

- (NSComparisonResult)compareAscending:(id)object
{
    NSComparisonResult result = NSOrderedSame;

    if ([self respondsToSelector:@selector(compare:)]) {
        result = (NSComparisonResult) [self performSelector:@selector(compare:)
                                                 withObject:object];
    }

    return result;
}

- (NSComparisonResult)compareDescending:(id)object
{
    NSComparisonResult result = NSOrderedSame;
    NSComparisonResult temp;

    if ([self respondsToSelector:@selector(compare:)]) {
        temp = (NSComparisonResult) [self performSelector:@selector(compare:)
                                               withObject:object];

        if (temp == NSOrderedAscending) {
            result = NSOrderedDescending;
        } else if (temp == NSOrderedDescending) {
            result = NSOrderedAscending;
        }
    }

    return result;
}

- (NSComparisonResult)compareCaseInsensitiveAscending:(id)object
{
    NSComparisonResult result = NSOrderedSame;

    if ([self respondsToSelector:@selector(caseInsensitiveCompare:)]) {
        result = (NSComparisonResult) [self performSelector:@selector(caseInsensitiveCompare:)
                                                 withObject:object];
    }

    return result;
}

- (NSComparisonResult)compareCaseInsensitiveDescending:(id)object
{
    NSComparisonResult result = NSOrderedSame;
    NSComparisonResult temp;

    if ([self respondsToSelector:@selector(caseInsensitiveCompare:)]) {
        temp = (NSComparisonResult) [self performSelector:@selector(caseInsensitiveCompare:)
                                               withObject:object];

        if (temp == NSOrderedAscending) {
            result = NSOrderedDescending;
        } else if (temp == NSOrderedDescending) {
            result = NSOrderedAscending;
        }
    }

    return result;
}

@end


@implementation NSArray (BDSortOrderingAdditions)

- (NSArray *)sortedArrayUsingKeyOrderArray:(NSArray *)keyOrderArray
{
    return [BDSortOrdering mergesortedArrayFromArray:self
                                  usingKeyOrderArray:keyOrderArray];
}

@end


@implementation NSMutableArray (BDSortOrderingAdditions)

- (void)sortUsingKeyOrderArray:(NSArray *)keyOrderArray
{
    [BDSortOrdering mergesortArrayInPlace:self
                       usingKeyOrderArray:keyOrderArray];
}

@end
