/*
    Copyright (c) 2002, 2003, bDistributed.com, Inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

    *   Redistributions of source code must retain the above
        copyright notice, this list of conditions and the following
        disclaimer.
    
    *   Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials
        provided with the distribution.
    
    *   Neither the name of bDistributed.com, Inc. nor the names of
        its contributors may be used to endorse or promote products
        derived from this software without specific prior written
        permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
    OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
    OF SUCH DAMAGE.
*/

#import "TestBDQualifier.h"

#import "BDQualifier.h"
#import "BDKeyValueQualifier.h"
#import "BDAndQualifier.h"


@interface SomeObject : NSObject
{
    NSString *_foo;
    NSString *_bar;
    NSString *_baz;
    double _price;
}

+ (id)someObjectWithPrice:(double)price;

- (id)init;
- (id)initWithPrice:(double)price;

- (void)dealloc;

- (NSString *)foo;
- (NSString *)bar;
- (double)price;
- (id)nada;
- (NSArray *)allValues;
- (SomeObject *)subObject;

@end


@implementation SomeObject

+ (id)someObjectWithPrice:(double)price
{
    return [[[self alloc] initWithPrice:price] autorelease];
}

- (id)init
{
    return [self initWithPrice:((double) 5.95)];
}

- (id)initWithPrice:(double)price
{
    if (self = [super init]) {
        _foo = @"foo";
        _bar = @"bar";
        _baz = @"baz";
        _price = price;
    }

    return self;
}

- (void)dealloc
{
    [super dealloc];
}

- (NSString *)foo
{
    return _foo;
}
- (NSString *)bar
{
    return _bar;
}
- (double)price
{
    return _price;
}
- (id)nada
{
    return nil;
}
- (NSArray *)allValues
{
    return [NSArray arrayWithObjects:_foo, _bar, _baz, [NSNumber numberWithDouble:_price], nil];
}
- (SomeObject *)subObject
{
    return [SomeObject someObjectWithPrice:(_price - ((double) 1.0))];
}

@end


@implementation TestBDQualifier

- (void)testKeyValueQualifierDescription
{
    BDQualifier *q;

    q = [[[BDKeyValueQualifier alloc] initWithKey:@"testKey"
                                 operatorSelector:BDQualifierOperatorEqual
                                            value:@"1"] autorelease];

    shouldBeEqual([q description], @"testKey = '1'");

    q = [[[BDKeyValueQualifier alloc] initWithKey:@"testKey"
                                 operatorSelector:BDQualifierOperatorEqual
                                            value:[NSNumber numberWithInt:1]]
        autorelease];

    shouldBeEqual([q description], @"testKey = 1");
}

- (void)testAndOrNotQualifierDescriptions
{
    BDQualifier *q;

    q = [BDQualifier qualifierWithQualifierFormat:
        @"(foo = 'foo' and bar = 'bar') or (not baz like 'f*')"];

    shouldBeEqual([q description],
                  @"((foo = 'foo') and (bar = 'bar')) or (not (baz like 'f*'))");
}

- (void)testQualifierParser1
{
    BDQualifier *q;

    q = [BDQualifier qualifierWithQualifierFormat:@"foo = 'bar'"];

    should([q isKindOfClass:[BDKeyValueQualifier class]]);

    shouldBeEqual([q description], @"foo = 'bar'");
}

- (void)testQualifierParser2
{
    BDQualifier *q;

    q = [BDQualifier qualifierWithQualifierFormat:@"baz like 'qu*'"];

    should([q isKindOfClass:[BDKeyValueQualifier class]]);

    shouldBeEqual([q description], @"baz like 'qu*'");
}

- (void)testQualifierEvaluation
{
    BDQualifier *q;
    SomeObject *object = [[[SomeObject alloc] init] autorelease];

    q = [BDQualifier qualifierWithQualifierFormat:@"foo = 'foo'"];

    should([q evaluateWithObject:object]);
}

- (void)testQualifierLike
{
    BDQualifier *q = [BDQualifier qualifierWithQualifierFormat:@"foo like 'f*'"];
    SomeObject *object = [[[SomeObject alloc] init] autorelease];

    should([q evaluateWithObject:object]);
}

- (void)testQualifierCaseInsensitiveLike
{
    BDQualifier *q = [BDQualifier qualifierWithQualifierFormat:@"foo caseInsensitiveLike 'FO?'"];
    SomeObject *object = [[[SomeObject alloc] init] autorelease];

    should([q evaluateWithObject:object]);
}

- (void)testQualifierUsingNumber
{
    BDQualifier *q;
    SomeObject *object = [[[SomeObject alloc] init] autorelease];

    q = [BDQualifier qualifierWithQualifierFormat:@"price == 5.95"];
    
    should([q evaluateWithObject:object]);
}

- (void)testFilteredArrayUsingQualifier
{
    BDQualifier *q;
    NSMutableArray *array;
    NSArray *filteredArray;

    array = [NSMutableArray arrayWithCapacity:5];

    [array addObject:[SomeObject someObjectWithPrice:1]];
    [array addObject:[SomeObject someObjectWithPrice:2]];
    [array addObject:[SomeObject someObjectWithPrice:3]];
    [array addObject:[SomeObject someObjectWithPrice:4]];
    [array addObject:[SomeObject someObjectWithPrice:5]];

    q = [BDQualifier qualifierWithQualifierFormat:@"price <= 3"];

    filteredArray = [array filteredArrayUsingQualifier:q];

    should(([[filteredArray objectAtIndex:0] price] == 1.0)
           && ([[filteredArray objectAtIndex:1] price] == 2.0)
           && ([[filteredArray objectAtIndex:2] price] == 3.0));
}

- (void)testCoding
{
    BDQualifier *q1;
    BDQualifier *q2;
    NSData *q1Data;
    SomeObject *object;

    q1 = [BDQualifier qualifierWithQualifierFormat:@"(baz like 'ba?') and (price = 3.95)"];

    q1Data = [NSArchiver archivedDataWithRootObject:q1];

    q2 = [NSUnarchiver unarchiveObjectWithData:q1Data];

    object = [SomeObject someObjectWithPrice:((double) 3.95)];

    should([q2 evaluateWithObject:object]);
}

- (void)testNull
{
    BDQualifier *q;
    SomeObject *object = [[[SomeObject alloc] init] autorelease];

    q = [BDQualifier qualifierWithQualifierFormat:@"nada = null"];

    should([q evaluateWithObject:object]);
}

- (void)testOneLevelKeyPaths
{
    BDQualifier *q;
    SomeObject *object = [[[SomeObject alloc] init] autorelease];

    q = [BDQualifier qualifierWithQualifierFormat:@"allValues.count = 4"];

    should([q evaluateWithObject:object]);
}

- (void)testTwoLevelKeyPaths
{
    BDQualifier *q;
    SomeObject *object = [[[SomeObject alloc] init] autorelease];

    q = [BDQualifier qualifierWithQualifierFormat:@"subObject.subObject.price = 3.95"];

    should([q evaluateWithObject:object]);
}

- (void)testParserFailureRecovery
{
    BDQualifier *q1, *q2;
    SomeObject *object = [[[SomeObject alloc] init] autorelease];

    q1 = [BDQualifier qualifierWithQualifierFormat:@"foo)"];

    q2 = [BDQualifier qualifierWithQualifierFormat:@"(foo = 'foo') and (bar = 'bar')"];

    should(q1 == nil);
    should([q2 evaluateWithObject:object]);
}

- (void)testVariable1
{
    BDQualifier *q1, *q2;
    SomeObject *object = [[[SomeObject alloc] init] autorelease];
    NSDictionary *bindings = [NSDictionary dictionaryWithObject:@"foo" forKey:@"x"];

    q1 = [BDQualifier qualifierWithQualifierFormat:@"foo = $x"];

    q2 = [q1 qualifierWithBindings:bindings
              requiresAllVariables:YES];

    should([q2 evaluateWithObject:object]);
}

- (void)testVariable2
{
    BDQualifier *q1, *q2;
    SomeObject *object = [[[SomeObject alloc] init] autorelease];
    NSDictionary *bindings = [NSDictionary dictionaryWithObjectsAndKeys:
        @"foo", @"x", @"ba?", @"y", nil];

    q1 = [BDQualifier qualifierWithQualifierFormat:@"(foo = $x) and (bar like $y)"];

    q2 = [q1 qualifierWithBindings:bindings
              requiresAllVariables:YES];

    should([q2 evaluateWithObject:object]);
}

- (void)testVariablePartial1
{
    BDQualifier *q1, *q2;
    NSDictionary *bindings = [NSDictionary dictionaryWithObjectsAndKeys:
        @"foo", @"x", nil];

    q1 = [BDQualifier qualifierWithQualifierFormat:@"(foo = $x) and (bar like $y)"];

    q2 = [q1 qualifierWithBindings:bindings
              requiresAllVariables:NO];

    shouldBeEqual([q2 description], @"foo = 'foo'");
}

- (void)testVariablePartial2
{
    BDQualifier *q1, *q2;
    NSDictionary *bindings = [NSDictionary dictionaryWithObjectsAndKeys:
        @"foo", @"x", @"ba?", @"y", nil];

    q1 = [BDQualifier qualifierWithQualifierFormat:@"(foo = $x) and (bar like $y) or (baz like $z)"];

    shouldBeEqual([q1 description], @"((foo = $x) and (bar like $y)) or (baz like $z)");

    q2 = [q1 qualifierWithBindings:bindings
              requiresAllVariables:NO];

    shouldBeEqual([q2 description], @"(foo = 'foo') and (bar like 'ba?')");
}

- (void)testParenthesization
{
    BDQualifier *q1, *q2, *q3;

    q1 = [BDQualifier qualifierWithQualifierFormat:@"(foo = 'foo') and (bar = 'bar') and (baz = 'baz')"];
    q2 = [BDQualifier qualifierWithQualifierFormat:@"(foo = 'foo') or (bar = 'bar') or (baz = 'baz')"];
    q3 = [BDQualifier qualifierWithQualifierFormat:@"(foo = 'foo') and (bar = 'bar') or (baz = 'baz')"];

    shouldBeEqual([q1 description], @"((foo = 'foo') and (bar = 'bar')) and (baz = 'baz')");
    shouldBeEqual([q2 description], @"((foo = 'foo') or (bar = 'bar')) or (baz = 'baz')");
    shouldBeEqual([q3 description], @"((foo = 'foo') and (bar = 'bar')) or (baz = 'baz')");
}

- (void)testBooleanQualifiers
{
    SomeObject *object = [[[SomeObject alloc] init] autorelease];
    BDQualifier *q1 = [BDQualifier qualifierWithQualifierFormat:@"*true*"];
    BDQualifier *q2 = [BDQualifier qualifierWithQualifierFormat:@"*false*"];
    BDQualifier *q3 = [BDQualifier qualifierWithQualifierFormat:@"*true* and *true*"];
    BDQualifier *q4 = [BDQualifier qualifierWithQualifierFormat:@"*true* and *false*"];

    shouldBeEqual([q1 description], @"*true*");
    shouldBeEqual([q2 description], @"*false*");
    shouldBeEqual([q3 description], @"(*true*) and (*true*)");
    shouldBeEqual([q4 description], @"(*true*) and (*false*)");

    should([q1 evaluateWithObject:object] == YES);
    should([q2 evaluateWithObject:object] == NO);
    should([q3 evaluateWithObject:object] == YES);
    should([q4 evaluateWithObject:object] == NO);
}

- (void)testQualifierCounts
{
    BDQualifier *q;

    q = [BDQualifier qualifierWithQualifierFormat:@"(foo = 'foo') and (bar = 'bar') and (baz = 'baz')"];

    should([q count] == 5);
}

- (void)testStringForOperatorSelector
{
    shouldBeEqual([BDQualifier stringForOperatorSelector:BDQualifierOperatorEqual],
                  @"=");
    shouldBeEqual([BDQualifier stringForOperatorSelector:BDQualifierOperatorNotEqual],
                  @"!=");
    shouldBeEqual([BDQualifier stringForOperatorSelector:BDQualifierOperatorLessThan],
                  @"<");
    shouldBeEqual([BDQualifier stringForOperatorSelector:BDQualifierOperatorGreaterThan],
                  @">");
    shouldBeEqual([BDQualifier stringForOperatorSelector:BDQualifierOperatorLessThanOrEqualTo],
                  @"<=");
    shouldBeEqual([BDQualifier stringForOperatorSelector:BDQualifierOperatorGreaterThanOrEqualTo],
                  @">=");
    shouldBeEqual([BDQualifier stringForOperatorSelector:BDQualifierOperatorContains],
                  @"contains");
    shouldBeEqual([BDQualifier stringForOperatorSelector:BDQualifierOperatorLike],
                  @"like");
    shouldBeEqual([BDQualifier stringForOperatorSelector:BDQualifierOperatorCaseInsensitiveLike],
                  @"caseInsensitiveLike");
}

@end
